extern crate actix_web;
extern crate chrono;
extern crate crypto;
extern crate jsonwebtoken as jwt;
#[macro_use]
extern crate log;
extern crate reqwest;
extern crate serde;
#[macro_use]
extern crate serde_derive;
extern crate serde_json;
extern crate uuid;

pub mod couchdb;
pub mod security;
pub mod web;

use std::collections::HashMap;
use std::default::Default;

use chrono::{DateTime, Utc};
use serde_json::Value;

#[derive(Serialize, Deserialize, Clone, Debug, Default)]
pub struct UserProfile {
    pub email: String,
}

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct User {
    #[serde(rename = "_id", skip_serializing_if = "Option::is_none")]
    id: Option<String>,
    #[serde(rename = "_rev", skip_serializing_if = "Option::is_none")]
    revision: Option<String>,
    #[serde(rename = "type", skip_serializing_if = "Option::is_none")]
    type_: Option<String>,

    pub username: String,
    #[serde(rename = "fullName")]
    pub full_name: String,
    pub password: String,
    pub profile: UserProfile,

    #[serde(skip_deserializing)]
    links: HashMap<String, Value>,
}

impl Default for User {
    fn default() -> User {
        User {
            id: None,
            revision: None,
            type_: Some(String::from("user")),
            username: Default::default(),
            full_name: Default::default(),
            password: Default::default(),
            profile: Default::default(),
            links: HashMap::new(),
        }
    }
}

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct Page {
    #[serde(rename = "_id", skip_serializing_if = "Option::is_none")]
    id: Option<String>,
    #[serde(rename = "_rev", skip_serializing_if = "Option::is_none")]
    revision: Option<String>,
    #[serde(rename = "type", skip_serializing_if = "Option::is_none")]
    type_: Option<String>,

    pub slug: String,
    pub title: String,
    pub content: String,
    #[serde(rename = "created", skip_deserializing)]
    created_at: Option<DateTime<Utc>>,

    #[serde(skip_deserializing)]
    links: HashMap<String, Value>,
}

impl Default for Page {
    fn default() -> Page {
        Page {
            id: None,
            revision: None,
            type_: Some(String::from("page")),
            title: Default::default(),
            slug: Default::default(),
            content: Default::default(),
            created_at: None,
            links: HashMap::new(),
        }
    }
}

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct MenuItems {
    pub title: String,
    pub url: String,
}

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct Site {
    #[serde(rename = "_id", skip_serializing_if = "Option::is_none")]
    id: Option<String>,
    #[serde(rename = "_rev", skip_serializing_if = "Option::is_none")]
    revision: Option<String>,
    #[serde(rename = "type", skip_serializing_if = "Option::is_none")]
    type_: Option<String>,

    pub name: String,
    #[serde(rename = "landingPage")]
    pub landing_page: String,
    pub header: String,
    pub footer: String,
    pub menu: Vec<MenuItems>,

    pub config: HashMap<String, Value>,

    #[serde(skip_deserializing)]
    links: HashMap<String, Value>,
}

impl Default for Site {
    fn default() -> Site {
        Site {
            id: None,
            revision: None,
            type_: Some(String::from("site")),
            name: Default::default(),
            landing_page: Default::default(),
            header: Default::default(),
            footer: Default::default(),
            menu: Default::default(),
            config: Default::default(),
            links: HashMap::new(),
        }
    }
}

pub trait RookeriesDomain {
    fn id(&self) -> Option<String>;
    fn remove_couchdb_details(&self) -> Self;
    fn add_self_links(&self, self_link: String) -> Self;
}

// TODO: Figure out how to set these as up default implementations.
impl RookeriesDomain for Page {
    fn id(&self) -> Option<String> {
        self.id.clone()
    }

    fn remove_couchdb_details(&self) -> Page {
        Page {
            id: None,
            revision: None,
            type_: None,
            ..self.clone()
        }
    }

    fn add_self_links(&self, self_link: String) -> Page {
        let mut links = HashMap::new();
        links.insert(String::from("self"), Value::String(self_link));

        Page {
            links,
            ..self.clone()
        }
    }
}

impl RookeriesDomain for Site {
    fn id(&self) -> Option<String> {
        self.id.clone()
    }

    fn remove_couchdb_details(&self) -> Site {
        Site {
            id: None,
            revision: None,
            type_: None,
            ..self.clone()
        }
    }

    fn add_self_links(&self, self_link: String) -> Site {
        let mut links = HashMap::new();
        links.insert(String::from("self"), Value::String(self_link));

        Site {
            links,
            ..self.clone()
        }
    }
}

impl RookeriesDomain for User {
    fn id(&self) -> Option<String> {
        self.id.clone()
    }

    fn remove_couchdb_details(&self) -> User {
        User {
            id: None,
            revision: None,
            type_: None,
            ..self.clone()
        }
    }

    fn add_self_links(&self, self_link: String) -> User {
        let mut links = HashMap::new();
        links.insert(String::from("self"), Value::String(self_link));

        User {
            links,
            ..self.clone()
        }
    }
}
