// TODO: Rework site management and testing.
// TODO: Rework modification to only allow changes to layout and children...
// TODO: Add a test if the site does not exist for some reason.

extern crate reqwest;
extern crate rookeries;
extern crate serde;
#[macro_use]
extern crate serde_json;

use reqwest::header::Authorization;
use reqwest::{Client, Response, StatusCode};
use rookeries::couchdb::Server;
use rookeries::{MenuItems, Site};
use serde_json::Value;

use std::collections::HashMap;

mod common;

fn auth_token_headers() -> String {
    let (user, password) = common::valid_user();
    let api_base_uri = common::api_base_url();

    let test_client = Client::new();
    let auth_url = api_base_uri.join("/auth").unwrap();
    let credentials = json!({"username": user.username, "password": password});

    let mut auth_response = test_client
        .post(auth_url)
        .json(&credentials)
        .send()
        .unwrap();

    let token = auth_response.json::<Value>().unwrap()["access_token"]
        .as_str()
        .unwrap()
        .to_string();
    format!("JWT {}", token)
}

fn test_site() -> Site {
    // TODO: Improve setup of test site.
    let mut test_config = HashMap::new();
    test_config.insert(
        String::from("tagline"),
        Value::from("Simple Site Scaffolding"),
    );
    test_config.insert(
        String::from("favicon"),
        Value::from("/static/images/mr-penguin-amber-favicon.ico"),
    );
    test_config.insert(
        String::from("logo"),
        Value::from("/static/images/mr-penguin-amber.svg"),
    );
    test_config.insert(String::from("footer"), Value::from("Footer here."));

    let mut test_site = Site::default();

    test_site.name = String::from("Test Site");
    test_site.menu = vec![
        MenuItems {
            title: String::from("Home"),
            url: String::from("/home"),
        },
        MenuItems {
            title: String::from("Contact"),
            url: String::from("/contact"),
        },
        MenuItems {
            title: String::from("Blog"),
            url: String::from("/blog"),
        },
    ];
    test_site.config = test_config;
    test_site.landing_page = String::from("/landing_page");
    test_site.header = String::from("Test Header");
    test_site.footer = String::from("Test Footer");

    let couchdb = Server::configure_from_env().unwrap();
    couchdb.update(test_site.clone()).unwrap();

    test_site
}

#[test]
fn test_visitor_can_access_site() {
    test_site();
    let api_base_uri = common::api_base_url();
    let client = Client::new();
    let response = client
        .get(api_base_uri.join("/api/site").unwrap())
        .send()
        .unwrap();

    assert_site_response(response);
}

#[test]
fn test_visitor_can_not_modify_site_using_proper_request() {
    test_site();

    let api_base_uri = common::api_base_url();
    let client = Client::new();
    let mut response = client
        .get(api_base_uri.join("/api/site").unwrap())
        .send()
        .unwrap();

    let original_site = response.json::<Site>().unwrap();

    let mut modified_site = original_site.clone();
    modified_site.name = String::from("Modified Test Site");

    let modified_site_response = client
        .put(api_base_uri.join("/api/site").unwrap())
        .json(&modified_site)
        .send()
        .unwrap();

    assert_bad_request_response(modified_site_response);
}

#[test]
fn test_authenticated_user_can_access_site() {
    test_site();
    let api_base_uri = common::api_base_url();
    let client = Client::new();
    let auth_token = auth_token_headers();
    let response = client
        .get(api_base_uri.join("/api/site").unwrap())
        .header(Authorization(auth_token))
        .send()
        .unwrap();

    assert_site_response(response);
}

#[test]
fn test_authenticated_user_cannot_modify_site_using_bad_request() {
    test_site();
    let api_base_uri = common::api_base_url();
    let client = Client::new();
    let auth_token = auth_token_headers();
    let response = client
        .put(api_base_uri.join("/api/site").unwrap())
        .body("")
        .header(Authorization(auth_token))
        .send()
        .unwrap();

    assert_bad_request_response(response);
}

#[test]
fn test_authenticated_user_cannot_modify_site_using_empty_request() {
    test_site();
    let api_base_uri = common::api_base_url();
    let client = Client::new();
    let auth_token = auth_token_headers();
    let response = client
        .put(api_base_uri.join("/api/site").unwrap())
        .json(&json!({}))
        .header(Authorization(auth_token))
        .send()
        .unwrap();

    assert_bad_request_response(response);
}

#[test]
fn test_authenticated_user_can_modify_site_using_proper_request() {
    test_site();

    let api_base_uri = common::api_base_url();
    let client = Client::new();
    let auth_token = auth_token_headers();
    let mut response = client
        .get(api_base_uri.join("/api/site").unwrap())
        .header(Authorization(auth_token.clone()))
        .send()
        .unwrap();

    let original_site = response.json::<Site>().unwrap();

    let mut modified_site = original_site.clone();
    modified_site.name = String::from("Modified Test Site");

    let modified_site_response = client
        .put(api_base_uri.join("/api/site").unwrap())
        .json(&modified_site)
        .header(Authorization(auth_token.clone()))
        .send()
        .unwrap();

    assert_updated_site_response(modified_site_response);

    // TODO: Remove once switch over to multi-site management...
    modified_site.name = String::from("Test Site");
    client
        .put(api_base_uri.join("/api/site").unwrap())
        .json(&modified_site)
        .header(Authorization(auth_token.clone()))
        .send()
        .unwrap();
}

fn assert_site_response(mut response: Response) {
    assert_eq!(response.status(), StatusCode::Ok);

    let expected_json = json!({
        "name": "Test Site",
        "links": {"self": "/api/site"},
        "config": {
            "tagline": "Simple Site Scaffolding",
            "favicon": "/static/images/mr-penguin-amber-favicon.ico",
            "logo": "/static/images/mr-penguin-amber.svg",
            "footer": "Footer here."
        },
        "landingPage": "/landing_page",
        "menu": [
            {"title": "Home", "url": "/home"},
            {"title": "Contact", "url": "/contact"},
            {"title": "Blog", "url": "/blog"},
        ],
        "header": "Test Header",
        "footer": "Test Footer",
    });

    let actual_json = response.json::<Value>().unwrap();
    assert_eq!(expected_json, actual_json);
}

fn assert_bad_request_response(mut response: Response) {
    let expected_response_json = json!({
        "error": {
            "status_code": 400,
            "message": "Bad request",
        }
    });

    assert_eq!(response.status(), StatusCode::BadRequest);
    let actual_json_response = response.json::<Value>().unwrap();
    assert_eq!(actual_json_response, expected_response_json);
}

fn assert_updated_site_response(mut response: Response) {
    assert_eq!(response.status(), StatusCode::Ok);

    let expected_json = json!({
        "name": "Modified Test Site",
        "links": {"self": "/api/site"},
        "config": {
            "tagline": "Simple Site Scaffolding",
            "favicon": "/static/images/mr-penguin-amber-favicon.ico",
            "logo": "/static/images/mr-penguin-amber.svg",
            "footer": "Footer here."
        },
        "landingPage": "/landing_page",
        "menu": [
            {"title": "Home", "url": "/home"},
            {"title": "Contact", "url": "/contact"},
            {"title": "Blog", "url": "/blog"},
        ],
        "header": "Test Header",
        "footer": "Test Footer",
    });

    let actual_json = response.json::<Value>().unwrap();
    assert_eq!(expected_json, actual_json);
}
